library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.libarray.all;

entity block_add is
Generic(N: integer:=4);-- 4 Adders = 8 numbers, X and Y frame size of 64*64
port(clk,rst_state,start:in std_logic;
	  done,accepted,zero_sad,done_count:out std_logic;
	  input_a: in std_logic;
	  input_ab: in input_array
);
end block_add;

architecture fulltree of block_add is

type int_sigs_ofz is array(0 to 3) of vect_array;
type cellshift is array (0 to 13) of std_logic_vector(1 downto 0);
type comp_s is array (0 to 5) of std_logic_vector(1 downto 0);
type int_sigs_ofz16 is array(0 to (N)-1)of std_logic_vector(1 downto 0);
type int_sigs_ofz32 is array(0 to (N/2)-1)of std_logic_vector(1 downto 0);
type blockStabs is (S0,S1,S2);
type blockSt is (S0,S1);
type mainSt is (S0,S1,S2,S3);

signal en,wr_wm,done13: std_logic;
signal out_wm : std_logic_vector(63 downto 0);
signal Sreg,state_next: blockSt;
signal Sreg1,state_next1:blockSt;
signal Sreg2,state_next2:blockSt;
signal Sreg3,state_next3:blockStabs;
signal Sreg4,state_next4:blockSt;
signal mainSreg,main_state_next:mainSt;

signal tocomp_s:comp_s;
signal out_z: int_sigs_ofz;
signal out_z16,comp_res16:int_sigs_ofz16;
signal out_z32,comp_res32:int_sigs_ofz32;
signal out_z64,out_ref:std_logic_vector(1 downto 0);
signal count_b: std_logic_vector(13 downto 0);
signal regdata,tempdata: cellshift;
signal comp_res64: std_logic_vector(1 downto 0);

signal done12,done14,done15,done18,done32,done0,lt,gt:std_logic;
signal done26,done29:std_logic;
signal rst_comp1,rst_comp2,rst_comp3:std_logic;
signal done_reg,accepted_reg,zero_sad_reg:std_logic;

signal ctr_reg, ctr_next: std_logic_vector(4 downto 0);
signal rst_sreg,rst_fsm,rst_abs,ctr_rst,rst:std_logic;
signal shift_en,load,wr_enable,rst_zsad: std_logic;

begin

-- Register storing SAD ref comparison value 
SADrefinit:entity work.regcell
port map(clk=>clk,rst=>rst_sreg,input_a=>tempdata(0),
			input_shift=>out_z64,
			load=>load,out_data=>regdata(0),shift_en=>shift_en);

SADref: for i in 1 to 13 generate
main_shifter:entity work.regcell
port map(clk=>clk,rst=>rst_sreg,input_a=>tempdata(i),
			input_shift=>regdata(i-1),
			load=>load,out_data=>regdata(i),shift_en=>shift_en);
end generate;

--Register inputs


--14 (2bit) storage registers to store last acceptable SAD: WriteEn control signal

process(clk)
begin
	
	if(clk'event and clk='1')then	
		if(rst_sreg='1')then
			tempdata <= (others=>"01");
		elsif(wr_enable='1')then
			tempdata <= regdata;
		end if;
	end if;	
end process;

-- end Storage registers

-- shifting registers for comparators
first_shift:entity work.d_ff2
port map(clk =>clk,rst=>rst,
			d=>regdata(13),q=>tocomp_s(0));

all_shift:for i in 1 to 5 generate
shifters_tocomp:entity work.d_ff2
port map(clk =>clk,rst=>rst,
			d=>tocomp_s(i-1),q=>tocomp_s(i));
end generate;

-- end of shifting registers

-- Absolute Value blocks
abs_val_0:for i in 0 to 15 generate
abs_valblock0:entity work.absvalue
port map(clk=>clk,rst_abs=>rst_abs,
			x_in=>input_ab(i),out_data=>out_z(0)(i));
end generate;

abs_val_1:for i in 0 to 15 generate
abs_valblock1:entity work.absvalue
port map(clk=>clk,rst_abs=>rst_abs,
			x_in=>input_ab(i+16),out_data=>out_z(1)(i));
end generate;

abs_val_2:for i in 0 to 15 generate
abs_valblock2:entity work.absvalue
port map(clk=>clk,rst_abs=>rst_abs,
			x_in=>input_ab(i+32),out_data=>out_z(2)(i));
end generate;

abs_val_3:for i in 0 to 15 generate
abs_valblock3:entity work.absvalue
port map(clk=>clk,rst_abs=>rst_abs,
			x_in=>input_ab(i+48),out_data=>out_z(3)(i));
end generate;

-- Adders Generation -- Online delay of 12 for 1 output + number of input digits for full output
-- after 12 cycles I get the first output at the begining of the 13th cycle
adderslvl_16:for i in 0 to N-1 generate
add8_col1:entity work.add_tree
port map(clk=>clk,rst=>rst,
			data_in=>out_z(i),z=>out_z16(i));
end generate;

addreslvl_32:for i in 0 to (N/2)-1 generate -- Online delay of 3
add16:entity work.sD_adder
port map(clk=>clk,rst=>rst,x=>out_z16(i),
			y=>out_z16(i+(N/2)),z=>out_z32(i));
end generate;
			
add64:entity work.sD_adder -- Online delay of 3
port map(clk=>clk,rst=>rst,x=>out_z32(0),
			y=>out_z32(1),z=>out_z64);

--First line comparators
Complvl16:for i in 0 to N-1 generate
comp16:entity work.olcomp
port map(clk=>clk,rst=>rst_comp1,
			a_in=>out_z16(i),b_in=>regdata(13),ltgt=>comp_res16(i));	--regdata(13)
end generate;	

--Second line comparators
Complvl32:for i in 0 to (N/2)-1 generate
comp32:entity work.olcomp
port map(clk=>clk,rst=>rst_comp2,
			a_in=>out_z32(i),b_in=>tocomp_s(2),ltgt=>comp_res32(i));	--tocomp_s(2)
end generate;

--Last output Comparator			
ola_comp:entity work.olcomp
port map(clk=>clk,rst=>rst_comp3,
			a_in=>out_z64,b_in=>tocomp_s(5),ltgt=>comp_res64);			--tocomp_s(5)
			
-- Watermark circuit
wmcircuit1:for i in 0 to 13 generate
wmcircuit:entity work.storecell
port map(clk=>clk,rst=>rst,input_a=>input_ab(i)(0),lsb_a=>count_b(i),
			en=>en,wr_wm=>wr_wm,out_wm=>out_wm(i));
end generate;			

wmcircuit2:for i in 14 to 63 generate
wmcircuit2:entity work.storecell
port map(clk=>clk,rst=>rst,input_a=>input_ab(i)(0),lsb_a=>input_ab(i)(0),
			en=>en,wr_wm=>wr_wm,out_wm=>out_wm(i));
end generate;	
-- Main State Machine and Process
process (clk)
begin
	if(clk'event and clk='1')then
		if (rst_state='1')then
			mainSreg <= S0;
			done <='0';
			accepted <='0';
		else 
			mainSreg <= main_state_next;	
			done <= done_reg;
			accepted <= accepted_reg;
		end if;
	end if;	
end process;

process(done12,done32,gt,start,mainSreg)
begin

		main_state_next <= mainSreg;
		rst<='0';
		rst_fsm<='0';
		rst_sreg<='0';
		load<='0';
		ctr_rst<='0';
		wr_enable<='0';
		accepted_reg<='0';
		done_reg<='0';
		shift_en <= '0';		
		wr_wm <='0';
		
		case mainSreg is		
			when S0 =>		
				if(start='0')then
					main_state_next<=S0;
					rst<='1';
					rst_fsm<='1';
					rst_sreg<='1';
					ctr_rst<='1';
				else
					main_state_next<=S1;
					load<='1';
					shift_en<='1';
				end if;
				
			when S1 =>
				
				if(done12='1')then
					load<='1';
					shift_en<='1';
					
				elsif(gt='1')then
					main_state_next<=S3;
					rst<='1';
					rst_fsm<='1';
					shift_en<='1';
					ctr_rst<='1';
					done_reg<='1';
					
				elsif(done32='1')then
					main_state_next<=S2;
					shift_en<='1';
					ctr_rst<='1';
					rst<='1';
					rst_fsm<='1';
					done_reg<='1';
					accepted_reg<='1';
					wr_wm<='1';
				else
					main_state_next<=S1;
					shift_en<='1';		
				end if;		
				
			when S2 =>
					main_state_next<=S1;
					wr_enable<='1';	
					ctr_rst<='1';	
					rst<='1';
					rst_fsm<='1';
					
					
			when S3 =>
					main_state_next<=S1;
					rst<='1';
					rst_fsm<='1';
					shift_en<='1';
					ctr_rst<='1';
					
					
			when others => null;
	end case;
end process;
			

-- Processes and State Machines, control units		
process (clk)
begin
	if(clk'event and clk='1')then
		if (rst_fsm='1')then
			Sreg<=S0;
			Sreg1<=S0;
			Sreg2<=S0;
			Sreg3<=S0;
		else 
			Sreg <= state_next;
			Sreg1 <= state_next1;
			Sreg2 <= state_next2;
			Sreg3 <= state_next3;

		end if;
	end if;	
end process;
	
-- Process for Zero SAD check
process (clk)
begin
	if(clk'event and clk='1')then
		if (rst_zsad='1')then
			Sreg4<=S0;
			zero_sad <= '1';
		else 
			zero_sad <= zero_sad_reg;
			Sreg4 <= state_next4;
		end if;
	end if;	
end process;				
	
-- Comparator 1 State Machine	
process (done12,done26,Sreg)
begin
		state_next <= Sreg;	
		rst_comp1 <= '0';
		
	case Sreg is		
	
			when S0 =>		
				if(done12='0')then
					state_next<=S0;
					rst_comp1<='1';
				else
					state_next<=S1;
					rst_comp1<='0';
				end if;
				
			when S1 =>
				if(done26='0')then
					state_next<=S1;
					rst_comp1<='0';
				else
					state_next<=S0;
					rst_comp1<='1';
				end if;				
		when others => null;
	end case;
end process;

-- Comparator 2 State Machine	
process (done15,done29,Sreg1)
begin
		state_next1 <= Sreg1;	
		rst_comp2 <= '0';		
	case Sreg1 is			
			when S0 =>		
				if(done15='0')then
					state_next1<=S0;
					rst_comp2<='1';
				else
					state_next1<=S1;
					rst_comp2<='0';
				end if;
				
			when S1 =>
				if(done29='0')then
					state_next1<=S1;
					rst_comp2<='0';
				else
					state_next1<=S0;
					rst_comp2<='1';
				end if;				
		when others => null;
	end case;
end process;

-- Comparator 3 State Machine	
process (done18,done32,Sreg2)
begin
		state_next2 <= Sreg2;	
		rst_comp3 <= '0';
		rst_zsad <= '0';
	case Sreg2 is			
			when S0 =>		
				if(done18='0')then
					state_next2<=S0;
					rst_comp3<='1';
				else
					state_next2<=S1;
					rst_comp3<='0';
					rst_zsad <='1';
				end if;
				
			when S1 =>
				if(done32='0')then
					state_next2<=S1;
					rst_comp3<='0';
				else
					state_next2<=S0;
					rst_comp3<='1';
				end if;				
		when others => null;
	end case;
end process;

-- SAD Zero Check
process (Sreg4)
begin
		state_next4 <= Sreg4;	
		zero_sad_reg <= '0';	
		
	case Sreg4 is			
			when S0 =>		
				if(out_z64="00" or out_z64="11")then
					state_next4<=S0;
					zero_sad_reg<='1';
				else
					state_next4<=S1;
					zero_sad_reg<='0';
				end if;
				
			when S1 =>
				state_next4<=S1;
				zero_sad_reg<='0';
			
			when others => null;
	end case;
end process;

-- Absolute Value Block State Machine	
process (done0,done14,done13,Sreg3)
begin
		state_next3 <= Sreg3;	
		rst_abs <= '0';	
		en <='0';
	case Sreg3 is			
			when S0 =>		
				if(done0='0')then
					state_next3<=S0;
					rst_abs<='1';
				else
					state_next3<=S1;
					rst_abs<='0';
					en <='1';
				end if;
				
			when S1 =>
				if(done13='1')then
					en<='0';
					state_next3<=S2;
				else
					en<='1';
				end if;
				
			when S2 =>		
				state_next3<=S0;
				rst_abs<='1';
					
				
		when others => null;
	end case;
end process;

-- Clock process and counter	
process(clk)
begin	
	if (clk'event and clk = '1') then
		if (ctr_rst = '1') then
			ctr_reg <= (others => '0');
		else
			ctr_reg <= ctr_next;	
		end if;	
	end if;
end process;

-- ltgt is the result with 10 A<B and 01 A>B
-- at counter 01100(12) return rst_abs to 1
--Counter outputs

--out_res <= out_z64;
ctr_next <= ctr_reg + 1;
done0  <= '1' when ctr_reg = "00101" else '0';
done12 <= '1' when ctr_reg = "01011" else '0';
done13 <= '1' when ctr_reg = "01100" else '0';
done14 <= '1' when ctr_reg = "01101" else '0';
done15 <= '1' when ctr_reg = "01110" else '0';
done18 <= '1' when ctr_reg = "10001" else '0';
done26 <= '1' when ctr_reg = "11001" else '0';
done29 <= '1' when ctr_reg = "11100" else '0';
done32 <= '1' when ctr_reg = "11111" else '0';

done_count <= done32;
--lt <= comp_res16(0)(1) or comp_res16(1)(1) or comp_res16(2)(1) or comp_res16(3)(1)
	--	or comp_res32(0)(1) or comp_res32(1)(1) or comp_res64(1);		
gt 	<= comp_res16(0)(0) or comp_res16(1)(0) or comp_res16(2)(0) or comp_res16(3)(0) or
         comp_res32(0)(0) or comp_res32(1)(0) or comp_res64(0);		

end fulltree;