library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.libarray.all;

-- Comparison Formula: R(current)=2*R(previous)+(A-B)
-- ltgt is the result with 10 A<B (diff is -2) and 01 A>B (Diff is 2)

entity olcomp is
port (clk,rst:in std_logic;
		a_in,b_in: in std_logic_vector(1 downto 0);
		ltgt: out std_logic_vector(1 downto 0)
		);	
end olcomp;

architecture online_compare of olcomp is
type compStates is (S1,S2,S3,S4,S5);
signal Sreg,state_next: compStates;
signal ltgt_reg: std_logic_vector(1 downto 0);
signal ab_concat: std_logic_vector(3 downto 0);
signal ab_comp: std_logic_vector(4 downto 0);
begin


ab_concat <= (a_in & b_in);
	
process(clk)
begin		
	if(clk'event and clk='1')then
		  if(rst='1')then
				Sreg <= S1;
				ltgt <= "00";
			else	
				ltgt <= ltgt_reg;		
				Sreg <= state_next;
			end if;	
	end if;		
end process;
  
process(ab_concat)
begin
	case (ab_concat) is
		when "0000" => ab_comp<="10000";
		when "0001" => ab_comp<="00100";								--ab_comp  10000 = 0
		when "0010" => ab_comp<="00010";								--ab_comp  01000 = -2
		when "0011" => ab_comp<="00010";								--ab_comp  00100 = -1
																				--ab_comp  00010 = 1
		when "0100" => ab_comp<="00010";								--ab_comp  00001 = 2
		when "0101" => ab_comp<="10000";
		when "0110" => ab_comp<="00001";
		when "0111" => ab_comp<="00010";
		
		when "1000" => ab_comp<="00100";
		when "1001" => ab_comp<="01000";
		when "1010" => ab_comp<="10000";
		when "1011" => ab_comp<="00100";
		
		when "1100" => ab_comp<="10000";
		when "1101" => ab_comp<="00100";
		when "1110" => ab_comp<="00010";
		when "1111" => ab_comp<="10000";
		
		when others => ab_comp<="10000";
	end case;	
end process;


process(Sreg,ab_comp)
-- S1 initial state   R=0
-- S2 second  state   R=-1
-- S3 thrid   state   R<-1 A<B
-- S4 fourth  state   R=1
-- S5 fifth   state   R=2 A>B		
begin
			state_next <= Sreg;
			ltgt_reg <= "00";
			
			case Sreg is			
			
				when S1 =>
					if(ab_comp(0)='1')then
					-- Result is 2 A>B	
						state_next<=S5;
						ltgt_reg<="01";
					end if;
					
					if(ab_comp(1)='1')then
					--Result is 1 
						state_next<=S4;
					end if;
					
					if(ab_comp(2)='1')then
					--Result is -1 
						state_next<=S2;
					end if;
					
					if(ab_comp(3)='1')then
					--Result is -2 A<B 
						state_next<=S3;
						ltgt_reg<="10";
					end if;
					
					if(ab_comp(4)='1')then
					--Result is 0
						state_next<=S1;
					end if;
					
				when S2 =>
					if(ab_comp(0)='1')then
					-- Result is 2 A>B	
						state_next <= S1;
					end if;
					
					if(ab_comp(1)='1')then
					--Result is 1 
						state_next <= S2;
					end if;
					
					if(ab_comp(2)='1' or ab_comp(4)='1' or ab_comp(3)='1')then
					--Result is -1,0 or -2
						state_next <= S3;
						ltgt_reg <= "10";
					end if;
				
				when S3 =>
					state_next <= S3;
					ltgt_reg <= "10";
					
				when S4 =>
					if(ab_comp(0)='1' or ab_comp(1)='1' or ab_comp(4)='1')then
					-- Result is 2 A>B or 0 or 1
						state_next<=S5;
						ltgt_reg<="01";
					end if;
					
					if(ab_comp(2)='1')then
					--Result is -1 
						state_next<=S4;
					end if;
					
					if(ab_comp(3)='1')then
					--Result is -2 A<B 
						state_next<=S1;
						ltgt_reg<="10";
					end if;
					
				when S5 =>
					state_next <= S5;
					ltgt_reg <= "01";
					
				 when others => null;
		end case;
end process;	
		
end online_compare;